import json
from .Basic import Basic

class Json(Basic):

	def __init__(self, params):
		"""
		Possible fields are: [
			'msg',
			'levelno', 'levelname',
			'module',
			'pathname',
			'filename',
			'lineno',
			'funcName',
			'exc_info', 'exc_text',
			'stack_info',
			'created', 'msecs', 'relativeCreated',
			'thread', 'threadName',
			'process', 'processName',
			'signature', 'timestamp', 'levelLetters', 'fileLine'
		]
		"""
		super().__init__(params)

		defaultFields = {"signature", "timestamp", "levelno", "fileLine", "msg",
		                 "process", "processName", "thread", "threadName"}

		self.fields = getattr(params, "fields", defaultFields) | getattr(params, "extraFields", set())


	def stringify(self, record):
		subRecord = {f: record.__dict__[f] for f in self.fields}
		subRecord.update(self.context)
		subRecord = {key: subRecord[key] for key in sorted(subRecord.keys())}
		return json.dumps(subRecord, ensure_ascii = False)

	@staticmethod
	def extractSignature(line):
		return json.loads(line)["signature"]




import logging
import secrets
import hmac
from types import SimpleNamespace
from contextlib import contextmanager

from logchain import formatters


VerbosityToLevel = {
	0: logging.ERROR,
	1: logging.WARNING,
	2: logging.INFO,
	3: logging.DEBUG,
}


class Result(SimpleNamespace):
	"""
	Class for reporting rich errors.
	Where a mere container would have been evaluated True,
	`Result` is evaluated like its `valid` argument.
	Extra information can be added as named arguments.
	"""
	def __init__(self, valid: bool, **kwargs):
		super().__init__(**kwargs)
		self.valid = valid

	def __bool__(self):
		return self.valid


class LogChainer:
	"""
	Entrypoint for initializing the log chain.
	"""

	def __init__(self, args = {}, **kwargs):
		"""
		@param args: dict
		"""

		defaultParams = {
			"stream": None,
			"formatterCls": formatters.Basic,
			"secret": secrets.token_urlsafe(128),
			"seed": secrets.token_urlsafe(),
			"format": "%(timestamp)s %(levelLetters)s %(fileLine)-15s %(funcName)-15s %(message)-60s |%(signature)s",
			"verbosity": 0,
			"timestampUtc": False,
			"timestampFmt": "iso", # or strftime
			"timestampPrecision": "milliseconds"
		}
		self.params = SimpleNamespace(**{**defaultParams, **args, **kwargs})
		self.formatter = self.params.formatterCls(self.params)


	def initLogging(self):
		aLevel = VerbosityToLevel.get(self.params.verbosity, logging.DEBUG)

		logging.basicConfig(level = aLevel, stream = self.params.stream)

		handler = logging.getLogger().handlers[0]
		handler.setFormatter(self.formatter)


	def verify(self, iLogChain):
		"""
		Check the consistency of a log chain with the secret.
		In case of failure, a tuple is returned with
		- check result: False
		- last consistent line
		- first inconsistent line
		"""
		extractFunc = getattr(self.params, "extractSignature", self.formatter.extractSignature)

		# TODO, verify the last line somehow
		for prevLine, line in zip(iLogChain, iLogChain[1:]):
			aStoredSign = extractFunc(line)
			aComputedSign = self.formatter.sign(prevLine, self.params.secret)

			isValid = hmac.compare_digest(aStoredSign, aComputedSign)

			if not isValid:
				#print("Inconsistency between lines:\nOK> %s\nKO> %s" % (prevLine, line))
				return Result(False, prevLine = prevLine, line = line)
		return Result(True)


	def setField(self, **kwargs):
		"""
		Adds contextual data to the log record in the form of key = value
		Remove a key by setting it to `None`.
		"""
		self.formatter.setField(**kwargs)

	@contextmanager
	def managedField(self, **kwargs):
		try:
			previousCtx = self.formatter.setField(**kwargs)
			yield
		finally:
			self.formatter.context = previousCtx

def stopLogging():
	"""
	Cleanup if needed, inspired from the library source code.
	"""
	logger = logging.getLogger()
	for h in logger.handlers[:]:
		logger.removeHandler(h)
		h.close()



import logging
from datetime import datetime, timezone

import hashlib
import hmac


class Basic(logging.Formatter):
	"""
	Base model for all custom formatters.
	"""
	def __init__(self, params):
		self.secret = params.secret
		self.prevLine = params.seed
		#self.signTrunc = getattr(params, "signTrunc", 16)

		# Have a check for presence of signature ?
		logging.Formatter.__init__(self, fmt = params.format)
		self.timestampFmt = params.timestampFmt
		self.timestampUtc = params.timestampUtc
		self.timestampPrecision = params.timestampPrecision
		self.context = {}


	def format(self, record):
		"""
		Shared method where the block-chain part is coded.
		"""
		record.signature = Basic.sign(self.prevLine, self.secret)
		record.levelLetters = record.levelname[:4]
		record.fileLine = record.filename  + ':' + str(record.lineno)
		record.timestamp = self.makeTimestamp(record)

		logLine = self.stringify(record)
		self.prevLine = logLine
		return logLine


	def makeTimestamp(self, record):
		"""
		Because formatTime called internally by logging.Formatter.format,
		we prefer a custom timestamp generation + an explicit `timestamp` name.
		We can customize the timezone, the decimal separator and number of digits.
		"""
		ts = datetime.fromtimestamp(record.created)

		if self.timestampUtc:
			ts = ts.astimezone(tz = timezone.utc).replace(tzinfo = None)

		if self.timestampFmt == "iso":
			return ts.isoformat(sep = ' ', timespec = self.timestampPrecision)
		else:
			# TODO, apply only if %f present
			aTrunc = 3 if self.timestampPrecision == "milliseconds" else None
			return ts.strftime(self.timestampFmt)[:aTrunc]


	def stringify(self, record):
		"""
		Specific method to transform the recod to string.
		"""
		return logging.Formatter.format(self, record)

	@staticmethod
	def sign(iMessageStr, iSecretStr, iLength = 16):
		"""
		Generates a truncated signature of the input message.
		@param iLength: controls the size of the signature, set it to None for full length.
		"""
		msg = bytes(iMessageStr, "utf-8")
		key = bytes(iSecretStr,  "utf-8")
		return hmac.new(key, msg, hashlib.sha256).hexdigest()[:iLength]

	@staticmethod
	def extractSignature(line):
		return line[line.rindex('|') + 1:]

	def setField(self, **kwargs):
		# Keep previous for scoped restoration
		previousCtx = {**self.context}
		self.context.update(kwargs)

		# Purge None values
		self.context = {k:v for k,v in self.context.items() if v is not None}
		return previousCtx



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['logchain', 'logchain.formatters']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'logchain',
    'version': '0.2.1',
    'description': 'Python Logging secured by blockchain',
    'long_description': '# Logchain\n\n[![pipeline status](https://gitlab.com/ggpack/logchain/badges/master/pipeline.svg)](https://gitlab.com/ggpack/logchain/-/commits/master)\n![Dependencies](https://img.shields.io/badge/dependencies-0-blue.svg)\n[![license](https://img.shields.io/badge/license-ISC-blue.svg)](https://gitlab.com/ggpack/logchain/-/blob/master/LICENSE)\n\n[![Web page](https://img.shields.io/badge/website-github.io/logchain-blue.svg)](https://gg-math.github.io/logchain)\n[![Package](https://img.shields.io/badge/PIP-logchain-blue.svg)](https://pypi.org/project/logchain)\n\nPython Logging based on blockchain 📜 ⛓️.\n\n## Logs get chained\nThe current log line contains the signature of the previous line with your secret.\n* detect lines deleted / lost\n* detect logs tampering\n\n## Philosophy\nThe package is intended to be a **lightweight** util for generating **incorruptible** logs.\n\nFor this pupose we rely as much as possible on standard packages: few dependencies, high quality.\n\nThe formatters are easy **extensible** by simply deriving from `Basic`.\n\n\n# Usage\n\n## Install\n``` bash\npip install logchain\n```\n\n## Choose your log type\n\nMany types of logs are supported out-of-the-box:\n- `Basic` raw text, relying on the standard formatter\n- `Json` structured log lines with static & dynamic fields\n- `CSV` work in progress\n\nYou can write a custom formatter in 20-ish lines.\n\n## Init once in main\n``` python\nfrom logchain import LogChainer\n\n# Initialize a default chainer.\ntheLogger = LogChainer()\n\n# Register the formatter to the logger.\ntheLogger.initLogging()\n```\n\nHave a look at [the comprehensive guide of constructor parameters](#constructor-parameters).\n\n## Use everywhere with python logging module\n``` python\nimport logging\n\nlogging.debug("My message")\nlogging.info("Some information")\n```\n\n## Check your logs integrity afterwards\n``` python\nfrom logchain import LogChainer\n\naLogChain = [\n\t"2020-03-30 13:38:00.782|0ec90b9839fdd964|TestChaining.py:20 test_logging_happy_case hello gg",\n\t"2020-03-30 13:38:00.782|2e3f1b4a7b946fb1|TestChaining.py:21 test_logging_happy_case voila1",\n\t"2020-03-30 13:38:00.782|10d1ab606618492a|TestChaining.py:22 test_logging_happy_case voila2",\n\t"2020-03-30 13:38:00.782|805757e144f4e385|TestChaining.py:23 test_logging_happy_case voila5",\n\t"2020-03-30 13:38:00.782|3bda90b5af77d3fe|TestChaining.py:24 test_logging_happy_case voila4"\n]\nresult = LogChainer.verify(aLogChain)\n\nif not result:\n\tprint("Last good line", result.prevLine)\n\tprint("First bad line", result.line)\nelse:\n\tprint("All right")\n```\n\n## Constructor parameters\n\nThey are passed as a dict and/or named arguments.\n``` python\nfrom logchain import LogChainer\n\ntheLogger = LogChainer(verbosity = 3, secret = "mySignatureKey")\n\nparams = {"verbosity": 3, "secret": "mySignatureKey"}\ntheLogger = LogChainer(params, timestampFmt = "iso")\n```\n\n| Param | Default value | Description |\n| ----- | ------------- | ----------- |\n| formatterCls | formatters.Basic | Type of logging to perform, raw text, json, custom |\n| format | see below | Placeholder string used by raw-text loggers |\n| secret | secrets.token_urlsafe(128) | Signature key to compute the line signature |\n| seed | secrets.token_urlsafe() | Random string to sign into the first log line |\n| timestampFmt | "iso" | iso for 8601 or `strftime` compatible placeholders |\n| timestampPrecision | "milliseconds" | `timespec` element used by [the datetime library](https://docs.python.org/3/library/datetime.html#datetime.datetime.isoformat) |\n| timestampUtc | False | Transform the timestamp to its value in UTC |\n| stream | cout | Where the logs are sent, file/console/custom stream |\n| verbosity | 0 | Number [0..5] mapped to a logging.level |\n\nThe default format is `%(timestamp)s %(levelLetters)s %(fileLine)-15s %(funcName)-15s %(message)-60s |%(signature)s`. It relies on some extra fields like the signature at its end.\n\n\n## Logchain extra logging fields\nWe enrich the standard logging record with some handy fields:\n\n| Name | Description |\n| ---- | ----------- |\n| fileLine | Widespread `filename:lineno` |\n| levelLetters | 4 first letters of logging level names: short and unambiguous |\n| signature | The digital signature of the previous line. Include it in all your lines to benefit from the chaining |\n| timestamp | Improved version of `asctime`, see below |\n\n\nThe `timestamp` field offers more flexibility than `asctime` in regards to:\n- the precision; can go up to the micro seconds (`msecs` cannot)\n- the decimal separator; you choose, \'.\' by default\n- utc or local timezone\n- customize the format only in one place: `timestampFmt`\n\n## Dynamic logging fields\nThe package is suitable for server/app logging which context changes from one transaction to another.\nHere is an example of setting contextual information throughout the lifecycle of an app:\n\n**App.py**\n``` python\nclass App:\n\tdef __init__(self, appName, logger):\n\t\tself.logger = logger\n\t\tself.logger.setField(appName = appName)\n\t\tlogging.info("Creating the app")\n\n\tdef handleTransaction(self, userId, callback):\n\t\twith self.logger.managedField(uId = userId, trxId = secrets.token_urlsafe(8)):\n\t\t\tcallback()\n\n\tdef close(self):\n\t\tlogging.info("Closing the app")\n```\n\n**Callbacks.py**\n``` python\n\t# The log chain in transparent for the callbacks\n\tdef callback1():\n\t\tlogging.warn("Something happened")\n\n\tdef callback2():\n\t\tlogging.info("Serving a resource")\n```\n\n**main.py**\n``` python\ndef main():\n\ttheLogger = logchain.LogChainer(formatterCls = logchain.formatters.Json)\n\ttheLogger.initLogging()\n\n\tapp = App("MyApp", theLogger)\n\tapp.handleTransaction("user1", callback1)\n\tapp.handleTransaction("user1", callback2)\n\tapp.close()\n```\n\nYou can either use:\n* `setField`: set a permanant field, remove it by setting it to `None`.\n* `managedField`: set a temporary field for the scope of the `context manager`.\n\n----\n\n# Contributing\n\n## Install\n[**The code is hosted on Gitlab 🦊**](https://gitlab.com/ggpack/logchain)\n\nSimply clone and submit pull requests.\n\n## Testing\nThe unit tests are located in the [test folder](https://gitlab.com/ggpack/logchain/-/blob/master/test)\nwhich contains the `__main__.py` entrypoint.\n\n``` bash\n# Run all\npython test\n\n# Get additional options\npython test --help\n```\n\n## Releasing\nThe process is triggered by a tag added to a commit. The tag must match the pattern `release_<VERSION>`\nand `VERSION` has to comply to **[semver](https://semver.org)**.\n\nA CI/CD job handles the new tag event and publishes the package to PYPI using\nusing to the awesome [Poetry tool](https://python-poetry.org):\n\n``` bash\npoetry version $VERSION\npoetry build\npoetry publish\n```\n',
    'author': 'Gg Math',
    'author_email': 'code@gmath.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gg-math.github.io/logchain',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.2,<4.0',
}


setup(**setup_kwargs)
